from keyweave.key_types import Key


num_1 = Key("num:1")
num_2 = Key("num:2")
num_3 = Key("num:3")
num_4 = Key("num:4")
num_5 = Key("num:5")
num_6 = Key("num:6")
num_7 = Key("num:7")
num_8 = Key("num:8")
num_9 = Key("num:9")
num_0 = Key("num:0")
num_enter = Key("num:enter")
num_plus = Key("num:+")
num_minus = Key("num:-")
num_asterisk = Key("num:*")
num_slash = Key("num:/")
num_dot = Key("num:.")

d_0 = Key("0")
d_1 = Key("1")
d_2 = Key("2")
d_3 = Key("3")
d_4 = Key("4")
d_5 = Key("5")
d_6 = Key("6")
d_7 = Key("7")
d_8 = Key("8")
d_9 = Key("9")
a = Key("a")
b = Key("b")
c = Key("c")
e = Key("e")
d = Key("d")
f = Key("f")
g = Key("g")
h = Key("h")
i = Key("i")
j = Key("j")
k = Key("k")
l = Key("l")
m = Key("m")
n = Key("n")
o = Key("o")
p = Key("p")
q = Key("q")
r = Key("r")
s = Key("s")
t = Key("t")
u = Key("u")
v = Key("v")
w = Key("w")
x = Key("x")
y = Key("y")
z = Key("z")

enter = Key("enter")
space = Key("space")
backspace = Key("backspace")
tab = Key("tab")
shift = Key("shift")
ctrl = Key("ctrl")
alt = Key("alt")
grave = Key("`")
minus = Key("-")
equal = Key("=")
plus = Key("+")
slash = Key("/")
backslash = Key("\\")
comma = Key(",")
dot = Key(".")
semicolon = Key(";")
quote = Key("'")
bracket_left = Key("[")
bracket_right = Key("]")

f1 = Key("f1")
f2 = Key("f2")
f3 = Key("f3")
f4 = Key("f4")
f5 = Key("f5")
f6 = Key("f6")
f7 = Key("f7")
f8 = Key("f8")
f9 = Key("f9")
f10 = Key("f10")
f11 = Key("f11")
f12 = Key("f12")

insert = Key("insert")
delete = Key("delete")
home = Key("home")
end = Key("end")
page_up = Key("page up")
page_down = Key("page down")

capslock = Key("capslock")
mouse_1 = Key("mouse:1")
mouse_2 = Key("mouse:2")
mouse_3 = Key("mouse:3")
